'use strict';

Object.defineProperty(exports, "__esModule", {
	value: true
});
exports.getData = exports.getDataFromFile = exports.handleReadFileErrors = exports.getPassphrase = exports.getPassphraseFromSource = exports.getPassphraseFromFile = exports.getPassphraseFromEnvVariable = exports.getPassphraseFromPrompt = exports.getStdIn = exports.splitSource = undefined;

var _fs = require('fs');

var _fs2 = _interopRequireDefault(_fs);

var _readline = require('readline');

var _readline2 = _interopRequireDefault(_readline);

var _inquirer = require('inquirer');

var _inquirer2 = _interopRequireDefault(_inquirer);

var _error = require('../error');

var _helpers = require('../helpers');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const capitalise = text => `${text.charAt(0).toUpperCase()}${text.slice(1)}`; /*
                                                                               * LiskHQ/lisk-commander
                                                                               * Copyright © 2017–2018 Lisk Foundation
                                                                               *
                                                                               * See the LICENSE file at the top-level directory of this distribution
                                                                               * for licensing information.
                                                                               *
                                                                               * Unless otherwise agreed in a custom licensing agreement with the Lisk Foundation,
                                                                               * no part of this software, including this file, may be copied, modified,
                                                                               * propagated, or distributed except according to the terms contained in the
                                                                               * LICENSE file.
                                                                               *
                                                                               * Removal or modification of this copyright notice is prohibited.
                                                                               *
                                                                               */


const getPassphraseVerificationFailError = displayName => `${capitalise(displayName)} was not successfully repeated.`;
const getPassphraseSourceTypeUnknownError = displayName => `${capitalise(displayName)} was provided with an unknown source type. Must be one of \`env\`, \`file\`, or \`stdin\`. Leave blank for prompt.`;
const getPassphraseEnvVariableNotSetError = displayName => `Environmental variable for ${displayName} not set.`;
const getFileDoesNotExistError = path => `File at ${path} does not exist.`;
const getFileUnreadableError = path => `File at ${path} could not be read.`;
const ERROR_DATA_MISSING = 'No data was provided.';
const ERROR_DATA_SOURCE = 'Unknown data source type.';
const DEFAULT_TIMEOUT = 100;

const splitSource = exports.splitSource = source => {
	const delimiter = ':';
	const sourceParts = source.split(delimiter);
	return {
		sourceType: sourceParts[0],
		sourceIdentifier: sourceParts.slice(1).join(delimiter)
	};
};

const getStdIn = exports.getStdIn = ({
	passphraseIsRequired,
	secondPassphraseIsRequired,
	passwordIsRequired,
	dataIsRequired
} = {}) => {
	const readFromStd = new Promise((resolve, reject) => {
		if (!(passphraseIsRequired || secondPassphraseIsRequired || passwordIsRequired || dataIsRequired)) {
			return resolve({});
		}
		const lines = [];
		const rl = _readline2.default.createInterface({ input: process.stdin });

		// Prevent readline hanging when command called with no input or piped
		const id = setTimeout(() => {
			clearTimeout(id);
			reject(new Error(`Timed out after ${DEFAULT_TIMEOUT} ms`));
		}, DEFAULT_TIMEOUT);

		const handleClose = () => {
			const passphraseIndex = 0;
			const passphrase = passphraseIsRequired ? lines[passphraseIndex] : null;

			const secondPassphraseIndex = passphraseIndex + (passphrase !== null);
			const secondPassphrase = secondPassphraseIsRequired ? lines[secondPassphraseIndex] : null;

			const passwordIndex = secondPassphraseIndex + (secondPassphrase !== null);
			const password = passwordIsRequired ? lines[passwordIndex] : null;

			const dataStartIndex = passwordIndex + (password !== null);
			const dataLines = lines.slice(dataStartIndex);

			return resolve({
				passphrase,
				secondPassphrase,
				password,
				data: dataLines.length ? dataLines.join('\n') : null
			});
		};

		return rl.on('line', line => lines.push(line)).on('close', handleClose);
	});
	return readFromStd;
};

const getPassphraseFromPrompt = exports.getPassphraseFromPrompt = async ({
	displayName,
	shouldRepeat
}) => {
	const questions = [{
		type: 'password',
		name: 'passphrase',
		message: `Please enter ${displayName}: `
	}];
	if (shouldRepeat) {
		questions.push({
			type: 'password',
			name: 'passphraseRepeat',
			message: `Please re-enter ${displayName}: `
		});
	}

	// Prompting user for additional input when piping commands causes error with stdin
	if (!(0, _helpers.stdoutIsTTY)() || !(0, _helpers.stdinIsTTY)()) {
		throw new Error(`Please enter ${displayName} using a flag when piping data.`);
	}

	const { passphrase, passphraseRepeat } = await _inquirer2.default.prompt(questions);
	if (shouldRepeat && passphrase !== passphraseRepeat) {
		throw new _error.ValidationError(getPassphraseVerificationFailError(displayName));
	}

	return passphrase;
};

const getPassphraseFromEnvVariable = exports.getPassphraseFromEnvVariable = async (key, displayName) => {
	const passphrase = process.env[key];
	if (!passphrase) {
		throw new _error.ValidationError(getPassphraseEnvVariableNotSetError(displayName));
	}
	return passphrase;
};

const getPassphraseFromFile = exports.getPassphraseFromFile = path => new Promise((resolve, reject) => {
	const stream = _fs2.default.createReadStream(path);
	const handleReadError = error => {
		stream.close();
		const { message } = error;

		if (message.match(/ENOENT/)) {
			return reject(new _error.FileSystemError(getFileDoesNotExistError(path)));
		}
		if (message.match(/EACCES/)) {
			return reject(new _error.FileSystemError(getFileUnreadableError(path)));
		}

		return reject(error);
	};
	const handleLine = line => {
		stream.close();
		resolve(line);
	};

	stream.on('error', handleReadError);

	_readline2.default.createInterface({ input: stream }).on('error', handleReadError).on('line', handleLine);
});

const getPassphraseFromSource = exports.getPassphraseFromSource = async (source, { displayName }) => {
	const { sourceType, sourceIdentifier } = splitSource(source);

	switch (sourceType) {
		case 'env':
			return getPassphraseFromEnvVariable(sourceIdentifier, displayName);
		case 'file':
			return getPassphraseFromFile(sourceIdentifier);
		case 'pass':
			return sourceIdentifier;
		default:
			throw new _error.ValidationError(getPassphraseSourceTypeUnknownError(displayName));
	}
};

const getPassphrase = exports.getPassphrase = async (passphraseSource, options) => {
	const optionsWithDefaults = Object.assign({ displayName: 'your secret passphrase' }, options);
	return passphraseSource && passphraseSource !== 'prompt' ? getPassphraseFromSource(passphraseSource, optionsWithDefaults) : getPassphraseFromPrompt(optionsWithDefaults);
};

const handleReadFileErrors = exports.handleReadFileErrors = path => error => {
	const { message } = error;
	if (message.match(/ENOENT/)) {
		throw new _error.FileSystemError(getFileDoesNotExistError(path));
	}
	if (message.match(/EACCES/)) {
		throw new _error.FileSystemError(getFileUnreadableError(path));
	}
	throw error;
};

const getDataFromFile = exports.getDataFromFile = async path => _fs2.default.readFileSync(path, 'utf8');

const getData = exports.getData = async source => {
	if (!source) {
		throw new _error.ValidationError(ERROR_DATA_MISSING);
	}

	const { sourceType, sourceIdentifier: path } = splitSource(source);

	if (sourceType !== 'file') {
		throw new _error.ValidationError(ERROR_DATA_SOURCE);
	}

	return getDataFromFile(path).catch(handleReadFileErrors(path));
};